@@ -104,14 +104,3 @@ except ImportError:
             return timedelta(0)
 
     utc = UTC()
-
-
-def hmset(pipe_or_connection, name, mapping):
-    # redis-py versions 3.5.0 and above accept a mapping parameter for hset
-    # This requires Redis server >= 4.0 so this is temporarily commented out
-    # and will be re-enabled at a later date
-    # try:
-    #    return pipe_or_connection.hset(name, mapping=mapping)
-    # earlier versions require hmset to be used
-    # except TypeError:
-    return pipe_or_connection.hmset(name, mapping)
