@@ -8,39 +8,19 @@ terminal colorizing code, originally by Georg Brandl.
 from __future__ import (absolute_import, division, print_function,
                         unicode_literals)
 
-import importlib
+import calendar
 import datetime
+import importlib
 import logging
-import os
+import numbers
 import sys
+try:
+    from collections.abc import Iterable
+except ImportError:
+    from collections import Iterable
 
-from .compat import is_python_version
-
-
-def gettermsize():
-    def ioctl_GWINSZ(fd):
-        try:
-            import fcntl
-            import struct
-            import termios
-            cr = struct.unpack('hh', fcntl.ioctl(fd, termios.TIOCGWINSZ, '1234'))
-        except:
-            return None
-        return cr
-    cr = ioctl_GWINSZ(0) or ioctl_GWINSZ(1) or ioctl_GWINSZ(2)
-    if not cr:
-        try:
-            fd = os.open(os.ctermid(), os.O_RDONLY)
-            cr = ioctl_GWINSZ(fd)
-            os.close(fd)
-        except:
-            pass
-    if not cr:
-        try:
-            cr = (os.environ['LINES'], os.environ['COLUMNS'])
-        except:
-            cr = (25, 80)
-    return int(cr[1]), int(cr[0])
+from .compat import as_text, is_python_version, string_types
+from .exceptions import TimeoutFormatError
 
 
 class _Colorizer(object):
@@ -85,35 +65,11 @@ class _Colorizer(object):
         return self.codes["reset"]
 
     def colorize(self, color_key, text):
-        if not sys.stdout.isatty():
+        if self.notty:
             return text
         else:
             return self.codes[color_key] + text + self.codes["reset"]
 
-    def ansiformat(self, attr, text):
-        """
-        Format ``text`` with a color and/or some attributes::
-
-            color       normal color
-            *color*     bold color
-            _color_     underlined color
-            +color+     blinking color
-        """
-        result = []
-        if attr[:1] == attr[-1:] == '+':
-            result.append(self.codes['blink'])
-            attr = attr[1:-1]
-        if attr[:1] == attr[-1:] == '*':
-            result.append(self.codes['bold'])
-            attr = attr[1:-1]
-        if attr[:1] == attr[-1:] == '_':
-            result.append(self.codes['underline'])
-            attr = attr[1:-1]
-        result.append(self.codes[attr])
-        result.append(text)
-        result.append(self.codes['reset'])
-        return ''.join(result)
-
 
 colorizer = _Colorizer()
 
@@ -145,10 +101,7 @@ class ColorizingStreamHandler(logging.StreamHandler):
 
     def __init__(self, exclude=None, *args, **kwargs):
         self.exclude = exclude
-        if is_python_version((2, 6)):
-            logging.StreamHandler.__init__(self, *args, **kwargs)
-        else:
-            super(ColorizingStreamHandler, self).__init__(*args, **kwargs)
+        super(ColorizingStreamHandler, self).__init__(*args, **kwargs)
 
     @property
     def is_tty(self):
@@ -180,16 +133,19 @@ def utcnow():
     return datetime.datetime.utcnow()
 
 
+_TIMESTAMP_FORMAT = '%Y-%m-%dT%H:%M:%S.%fZ'
+
+
 def utcformat(dt):
-    return dt.strftime(u'%Y-%m-%dT%H:%M:%SZ')
+    return dt.strftime(as_text(_TIMESTAMP_FORMAT))
 
 
 def utcparse(string):
     try:
-        return datetime.datetime.strptime(string, '%Y-%m-%dT%H:%M:%SZ')
+        return datetime.datetime.strptime(string, _TIMESTAMP_FORMAT)
     except ValueError:
-        # This catches RQ < 0.4 datetime format
-        return datetime.datetime.strptime(string, '%Y-%m-%dT%H:%M:%S.%f+00:00')
+        # This catches any jobs remain with old datetime format
+        return datetime.datetime.strptime(string, '%Y-%m-%dT%H:%M:%SZ')
 
 
 def first(iterable, default=None, key=None):
@@ -229,3 +185,66 @@ def first(iterable, default=None, key=None):
                 return el
 
     return default
+
+
+def is_nonstring_iterable(obj):
+    """Returns whether the obj is an iterable, but not a string"""
+    return isinstance(obj, Iterable) and not isinstance(obj, string_types)
+
+
+def ensure_list(obj):
+    """
+    When passed an iterable of objects, does nothing, otherwise, it returns
+    a list with just that object in it.
+    """
+    return obj if is_nonstring_iterable(obj) else [obj]
+
+
+def current_timestamp():
+    """Returns current UTC timestamp"""
+    return calendar.timegm(datetime.datetime.utcnow().utctimetuple())
+
+
+def enum(name, *sequential, **named):
+    values = dict(zip(sequential, range(len(sequential))), **named)
+
+    # NOTE: Yes, we *really* want to cast using str() here.
+    # On Python 2 type() requires a byte string (which is str() on Python 2).
+    # On Python 3 it does not matter, so we'll use str(), which acts as
+    # a no-op.
+    return type(str(name), (), values)
+
+
+def backend_class(holder, default_name, override=None):
+    """Get a backend class using its default attribute name or an override"""
+    if override is None:
+        return getattr(holder, default_name)
+    elif isinstance(override, string_types):
+        return import_attribute(override)
+    else:
+        return override
+
+
+def str_to_date(date_str):
+    if not date_str:
+        return
+    else:
+        return utcparse(date_str.decode())
+
+
+def parse_timeout(timeout):
+    """Transfer all kinds of timeout format to an integer representing seconds"""
+    if not isinstance(timeout, numbers.Integral) and timeout is not None:
+        try:
+            timeout = int(timeout)
+        except ValueError:
+            digit, unit = timeout[:-1], (timeout[-1:]).lower()
+            unit_second = {'d': 86400, 'h': 3600, 'm': 60, 's': 1}
+            try:
+                timeout = int(digit) * unit_second[unit]
+            except (ValueError, KeyError):
+                raise TimeoutFormatError('Timeout must be an integer or a string representing an integer, or '
+                                         'a string with format: digits + unit, unit can be "d", "h", "m", "s", '
+                                         'such as "1h", "23m".')
+
+    return timeout
